! In this module routines and functions necessary for the main are defined 
MODULE tools_module
    USE types_module
    IMPLICIT NONE  
    
CONTAINS
    
    FUNCTION ErrorFunction(inputpar,targets,states)
        ! This subroutine returns a discrepancy.
        ! The process passed to the routine is used to create a discrete
        ! approximation of the process based on a specific assignment of states.
        ! The discrete approximation is then tested against a number of targets
        ! The error is the % squared discrepancy between the
        ! targeted moments of the continuous process and the targeted moments of
        ! discrete process.
        REAL(dp), INTENT(IN), DIMENSION(9)      ::  inputpar
        REAL(dp), INTENT(IN), DIMENSION(9)      ::  targets
        REAL(dp), INTENT(IN), DIMENSION(:)      ::  states
        REAL(dp)                                ::  ErrorFunction
		! inputpar contains the paramenters to specify the NMAR
		! targets contains the values of the targeted moments
		! states is the state assignment that is being tested
		! errorfunction returns the discrepancy from the target for this
		! 	specific state assignment
        REAL(dp), DIMENSION(SIZE(states),SIZE(states))  ::  transition          
        REAL(dp), DIMENSION(SIZE(states))               ::  stationaryvec          
		REAL(dp), DIMENSION(size(states)-1)    			::  dif_vec
        REAL(dp), DIMENSION(9)                          ::  moments 
		REAL(dp), DIMENSION(9)                          ::  CriterionVec        
        REAL(dp)                                        ::  temp,temp1
        INTEGER                                         ::  i,j,n,flag,info
        n=SIZE(states)
        ErrorFunction=0.0_dp
        flag=0
		! the following is a penalty function necessary to ensure the states are
		! assigned in a monotonically increasing fashion
		! if this is not true a penalty is assigned and the states are sorted
        DO i=1,n
            DO j=1,n
                IF ( (i .GT. j) .AND. (states(i)-states(j) .LT. 0.0_dp) ) THEN
                    ErrorFunction=ErrorFunction+(states(i)-states(j))**(2.0_dp)*(10.0_dp**4.0_dp)
                    flag=1
                END IF
            END DO
        END DO
        IF (flag .EQ. 1) THEN
            CALL dlasrt('I',n,states,info)
        END IF
        ! compute the transition matrix based on the states and parameters
        CALL TransitionOneStep(inputpar,states,transition)
        ! a numerical check that the elements of the transitions make sense
        DO i=1,n
            DO j=1,n
                IF (transition(i,j) .GT. 1.0_dp) STOP 'gt'
                IF (transition(i,j) .LT. 0.0_dp) STOP 'lt'
            END DO
        END DO
        DO i=1,n
            IF (SUM(transition(i,:)) .LT. 0.99_dp) STOP 'not1'
        END DO
        ! based on the transition matrix and on the states we compute the
        ! relevant moments of the discrete process that we want to target
        moments=TransitionStates2Targets(transition,states)
        ! we compute the squared discrepancies
		CriterionVec=0.0_dp 
        ! CriterionVec(1)=((targets(1)-moments(1))/targets(1))**2.0*100.0_dp
        CriterionVec(2)=((targets(2)-moments(2))/targets(2))**2.0*100.0_dp
		CriterionVec(3)=((targets(3)-moments(3))/targets(3))**2.0*100.0_dp
        CriterionVec(4)=((targets(4)-moments(4))/targets(4))**2.0*100.0_dp
        ! CriterionVec(5)=((targets(5)-moments(5))/targets(5))**2.0*100.0_dp!*0.0_dp
        CriterionVec(6)=((targets(6)-moments(6))/targets(6))**2.0*100.0_dp           
		CriterionVec(7)=((targets(7)-moments(7))/targets(7))**2.0*100.0_dp   
		CriterionVec(8)=((targets(8)-moments(8))/targets(8))**2.0*100.0_dp   
        CriterionVec(9)=((targets(9)-moments(9))/targets(9))**2.0*100.0_dp   
        ! we sum the discrepancies and average them
        ErrorFunction=SUM(CriterionVec)/7.0_dp+ErrorFunction
        ! the following penalization ensures that the transition matrix has some desirable
        ! properties
		!
		! diagonal elements less than 0.995
		! this ensures that no state is absorbing, or close to
        DO i=1,n
            IF (transition(i,i)>0.995) THEN
                ErrorFunction=ErrorFunction+(transition(i,i)-0.995)**(2.0_dp)*(10.0_dp**7.0_dp)
            END IF
        END DO
		! from every state, there is a positive prob of going up or down
		! this ensures that the markov process is irreducible
        DO i=1,n
            IF (i .EQ. 1) THEN
                temp=sum(transition(i,i+1:n))
                ErrorFunction=ErrorFunction+MIN(temp-0.005_dp,0.0_sp)**(2.0_dp)*(10.0_dp**7.0_dp)
            ELSE IF (i .EQ. n) THEN
                temp=sum(transition(i,1:i-1))
                ErrorFunction=ErrorFunction+MIN(temp-0.005_dp,0.0_sp)**(2.0_dp)*(10.0_dp**7.0_dp)
            ELSE
                temp  = sum(transition(i,1:i-1))
                temp1 = sum(transition(i,i+1:n))
                ErrorFunction=ErrorFunction+MIN(temp-0.005_dp,0.0_sp)**(2.0_dp)*(10.0_dp**7.0_dp)
                ErrorFunction=ErrorFunction+MIN(temp1-0.005_dp,0.0_sp)**(2.0_dp)*(10.0_dp**7.0_dp)
            END IF
        END DO
		! Stationary probabilities are greater than 0.001
		! this ensures all the states are visited
		! this restriction can in fact be relaxed, especially to allow the MC to
		! match higher kurtosis
        stationaryvec=stationary(transition)
        temp=SUM(MIN(stationaryvec-0.001_dp,0.0_dp)**(2.0_dp)*(10.0_dp**7.0_dp))
        ErrorFunction=ErrorFunction+temp
		! single peaked stationary probabilities
		! this ensures single peakedness of the stationary distribution when such
		! a property is desirable 
		! if the NMAR is not single peaked this restriction should be relaxed or eliminated
		temp=0.0_dp
		dif_vec=0.0_dp
		i=MAXLOC(stationaryvec,1) !for single peaked stationary distribution
		IF(i .EQ. 1) THEN
			do j=1,n-1
				dif_vec(j)=stationaryvec(j)-stationaryvec(j+1) !We want this vector to be all positive
			end do
		ELSE IF(i .EQ. n) THEN
			do j=1,n-1
				dif_vec(j)=stationaryvec(j+1)-stationaryvec(j)
			end do
		ELSE
			do j=1,i-1
				dif_vec(j)=stationaryvec(j+1)-stationaryvec(j)
			end do
			do j=i,n-1
				dif_vec(j)=stationaryvec(j)-stationaryvec(j+1)
			end do
		END IF
		temp=SUM(MIN(dif_vec,0.0_dp)**(2.0_dp)*(10.0_dp**10.0_dp))
		! summing the discrepancy and penalties together
		ErrorFunction=ErrorFunction+temp
    END FUNCTION ErrorFunction

    FUNCTION PerError(inputpar,targets,states,detail)
		! this function computes the percentage discrepancy from the
		! targets
		! notice there are no penalty functions
        REAL(dp), INTENT(IN), DIMENSION(9)      ::  inputpar
        REAL(dp), INTENT(IN), DIMENSION(9)      ::  targets
        REAL(dp), INTENT(IN), DIMENSION(:)      ::  states
        REAL(dp)                                ::  PerError
        REAL(dp), INTENT(OUT), DIMENSION(9)     ::  detail
		!
        REAL(dp), DIMENSION(SIZE(states),SIZE(states))  ::  transition
        REAL(dp), DIMENSION(9)                          ::  moments
        REAL(dp), DIMENSION(9)                          ::  CriterionVec
		REAL(dp), DIMENSION(9)                          ::  whatmoments
        INTEGER                                         ::  n
        n=SIZE(states)
        PerError=0.0_dp
        CALL TransitionOneStep(inputpar,states,transition)
        moments=TransitionStates2Targets(transition,states)
		whatmoments=1.0_dp
		!
		whatmoments(1)=0.0_dp
		whatmoments(5)=0.0_dp
		!
		CriterionVec=0.0_dp
		CriterionVec(1)=moments(1)
		CriterionVec(5)=moments(5)
		!CriterionVec(3)=moments(3)
		!CriterionVec(7)=moments(7)
        !CriterionVec(1)=(moments(1)-targets(1))/targets(1)*100.0_dp 
        CriterionVec(2)=(moments(2)-targets(2))/targets(2)*100.0_dp!*0.0_dp
        CriterionVec(3)=(moments(3)-targets(3))/targets(3)*100.0_dp!*0.0_dp 
		CriterionVec(4)=(moments(4)-targets(4))/targets(4)*100.0_dp!*0.0_dp 
        !CriterionVec(5)=(moments(5)-targets(5))/targets(5)*100.0_dp!*0.0_dp
        CriterionVec(6)=(moments(6)-targets(6))/targets(6)*100.0_dp
        CriterionVec(7)=(moments(7)-targets(7))/targets(7)*100.0_dp
        CriterionVec(8)=(moments(8)-targets(8))/targets(8)*100.0_dp
        CriterionVec(9)=(moments(9)-targets(9))/targets(9)*100.0_dp
        detail=CriterionVec
        PerError=SUM(ABS(CriterionVec)*whatmoments)/SUM(whatmoments)
    END FUNCTION PerError
    
    FUNCTION ParToMom(inputpar)
        IMPLICIT NONE
        REAL(dp), DIMENSION(9), INTENT(IN) ::  inputpar
        REAL(dp), DIMENSION(9)             ::  ParToMom
        REAL(dp)                           ::  m1,m2,m3,s21,s22,s23,p1,p2,p3,rho
        REAL(dp)                           ::  e1,e2,e3,e4,Ee,Ve,Se,Ke,Ey,Vy,Sy,Ky
        !
        !  this function computes the moments of the AR process
        !  with normal mixture shocks, using analytic formulas
        !  the list of moments returned is as follows
        !
        ! 1 mean of the process
        ! 2 variance of the process
        ! 3 skewness of the process
        ! 4 kurtosis of the process
        ! 5 mean of the e
        ! 6 variance of the e
        ! 7 skewness of the e
        ! 8 kurtosis of the e
        ! 9 rho
        !
        !   the parameters are the following
        ! 1 mean1
        ! 2 variance1
        ! 3 mean2
        ! 4 variance2
        ! 5 mean3    
        ! 6 variance3
        ! 7 p1
        ! 8 p2
        ! 9 rho
        !
        m1  = inputpar(1)
        m2  = inputpar(3)
        m3  = inputpar(5)
        s21 = inputpar(2)
        s22 = inputpar(4)
        s23 = inputpar(6)
        p1  = inputpar(7)
        p2  = inputpar(8)
        p3  = 1.0_dp-p1-p2
        rho = inputpar(9)
		
        ParToMom(9) = rho
        e1 = p1*m1+p2*m2+p3*m3
        e2 = p1*(m1**2.0_dp+s21)+p2*(m2**2.0_dp+s22)+p3*(m3**2.0_dp+s23)
        e3 = p1*(m1**3.0_dp+3.0_dp*m1*s21)+p2*(m2**3.0_dp+3.0_dp*m2*s22)+p3*(m3**3.0_dp+3.0_dp*m3*s23)
        e4 = p1*(m1**4.0_dp+6.0_dp*m1**2.0_dp*s21+3.0_dp*s21**2.0_dp) &
           + p2*(m2**4.0_dp+6.0_dp*m2**2.0_dp*s22+3.0_dp*s22**2.0_dp) &
           + p3*(m3**4.0_dp+6.0_dp*m3**2.0_dp*s23+3.0_dp*s23**2.0_dp)
        Ee = e1
        Ve = e2-e1**2.0_dp
        Se = (e3-3.0_dp*e1*Ve-e1**3.0_dp)/(Ve**0.5_dp)**3.0_dp
        Ke = (e4-3.0_dp*e1**4.0_dp+6.0_dp*e2*e1**2.0_dp-4.0_dp*e1*e3)/Ve**2.0_dp
        Ey = e1/(1.0_dp-rho)
        Vy = Ve/(1.0_dp-rho**2.0_dp)
        Sy = Se*Ve**1.5_dp/((1.0_dp-rho**3.0_dp)*Vy**1.5_dp)
        Ky = 3.0_dp+Ve**2.0_dp*(Ke-3.0_dp)/(1.0_dp-rho**4.0_dp)/Vy**2.0_dp
        ParToMom(1)=Ey
        ParToMom(2)=Vy
        ParToMom(3)=Sy
        ParToMom(4)=Ky
        ParToMom(5)=Ee
        ParToMom(6)=Ve
        ParToMom(7)=Se
        ParToMom(8)=Ke
    END FUNCTION ParToMom
    
    SUBROUTINE DiffHist(t,s)
        REAL(dp), DIMENSION(:,:), INTENT(IN)                    ::  t
        REAL(dp), DIMENSION(:), INTENT(IN)                      ::  s
        REAL(dp), DIMENSION(SIZE(s)**2-SIZE(s)+1)               ::  points,hist
        INTEGER                                                 ::  i,j,N,temp
        REAL(dp), DIMENSION(SIZE(s),SIZE(s))                    ::  temp1,temp2
        REAL(dp), DIMENSION(SIZE(s))                            ::  ss
        CHARACTER(2)                                            ::  name
        write (name,'(I2)') SIZE(s)
        OPEN(UNIT=3,FILE='hist'//name)
        points=0.0
        hist=0.0
        N=SIZE(s)
        DO i=1,N
            temp1(i,:)=s
        END DO
        temp1=temp1-TRANSPOSE(temp1)
        ss=stationary(t)
        DO i=1,N
            temp2(i,:)=ss(i)*t(i,:)
        END DO
        temp=1
        DO i=1,N
            DO j=1,N
                IF (i .NE. j) THEN
                    points(temp)=temp1(i,j)
                    hist(temp)=temp2(i,j)
                    temp=temp+1
                END IF
            END DO
        END DO
        points(temp)=0.0_dp
        hist(temp)=1.0_dp-SUM(hist)
        WRITE(3,*)points
        WRITE(3,*)hist
        CLOSE(3)
    END SUBROUTINE DiffHist
                
    SUBROUTINE simulate(CumTransition,SimulatedProcess)
        IMPLICIT NONE
        REAL(dp), DIMENSION(:,:), INTENT(IN)  ::  CumTransition     ! the cumulated transition matrix
        REAL(dp), DIMENSION(:), INTENT(OUT)   ::  SimulatedProcess  ! the realizations of the continuous process
        ! this subroutine returns a vector of simulated data from a markov process characterized by states and transition probabilities
        INTEGER                                     ::  n                 ! number of states used for the discrete approximation
        INTEGER                                     ::  t1                ! number of simulations
        INTEGER                                     ::  i               ! counters
        REAL(dp), DIMENSION(SIZE(CumTransition,1))  ::  difference        ! housekeeping
        REAL(dp)                                    ::  prob
        !
        prob=rand(3456)
        n=SIZE(CumTransition,1)
        t1=SIZE(SimulatedProcess)
        ! we set the first simulated value
        SimulatedProcess(1)=FLOOR(n/2.0_dp)
        ! we make the values fluctuate according to the transition probabilities
        DO i=2,t1
            prob=rand()
            difference=CumTransition(INT(SimulatedProcess(i-1)),:)-prob
            SimulatedProcess(i)=MINLOC(difference,1,difference .GT. 0.0_dp)
        END DO
    END SUBROUTINE simulate
    
    FUNCTION cumulative(x)
        REAL(DP), DIMENSION(:,:), INTENT(IN)      :: x
        REAL(DP), DIMENSION(size(x,1),size(x,2))  :: cumulative
        ! this function returns the cumulative of a transition matrix x
        INTEGER                                   :: row,col
        INTEGER                                   :: counter1,counter2
        row=size(x,1)
        col=size(x,2)
        DO counter1=1,row
            DO counter2=1,col
                IF (counter2 .EQ. 1) THEN
                    cumulative(counter1,counter2)=x(counter1,counter2)
                ELSE
                    cumulative(counter1,counter2)=cumulative(counter1,counter2-1)+x(counter1,counter2)
                END IF
            END DO
        END DO
    END FUNCTION            

    SUBROUTINE draw_process(inputpar,process)
        IMPLICIT NONE
        REAL(dp), INTENT(IN), DIMENSION(9)      ::  inputpar
        REAL(dp), DIMENSION(:), INTENT(OUT)     ::  process
        ! draw_process returns process, a vector of realizations from a continuous stochastic process
        ! the continuous stochastic process in question is y_t=\rho*y_t-1+\epsilon
        ! where the innovation is a normal mixture described below
        ! this normal mixture is dependent on the parameters var1,var2,var3,mean3 passed to the subroutine
        INTEGER                                 ::  t       ! number of realizations of the continuous process
        INTEGER                                 ::  i       ! counter
        REAL(dp)                                ::  prob
        !
!         call seed(2222)
        prob=rand(2222)
        t=SIZE(process) 
        process=0.0_dp   
        ! this is the sequence of shocks from the normal mixure 
        DO i=1,t
            prob=rand()
            IF (prob .LE. inputpar(7)) THEN
                process(i)=randnorm()*SQRT(inputpar(2))+inputpar(1)
            ELSE IF (prob .LE. inputpar(7)+inputpar(8)) THEN
                process(i)=randnorm()*SQRT(inputpar(4))+inputpar(3)
            ELSE
                process(i)=randnorm()*SQRT(inputpar(6))+inputpar(5)
            END IF
        END DO
        ! here we combine the shocks to obtain the process
        DO i=2,t
            process(i)=process(i-1)*inputpar(9)+process(i)
        END DO
    END SUBROUTINE draw_process
    
    SUBROUTINE TransitionOneStep(inputpar,states,transition)
        IMPLICIT NONE
        REAL(dp), INTENT(IN), DIMENSION(9)                          ::  inputpar
        REAL(dp), DIMENSION(:), INTENT(IN)                          ::  states
        REAL(dp), DIMENSION(size(states),size(states)), INTENT(OUT) ::  transition
        !
        REAL(dp)                                                    ::  m1,s1,m2,s2,m3,s3,p1,p2,rho
        REAL(dp), DIMENSION(SIZE(states)-1)                         ::  nodes
        INTEGER                                                     ::  i,j,n
        !
        m1=inputpar(1)          ! mean1
        s1=inputpar(2)**0.5_dp  ! sd1
        m2=inputpar(3)          ! mean2
        s2=inputpar(4)**0.5_dp  ! sd2
        m3=inputpar(5)          ! mean3    
        s3=inputpar(6)**0.5_dp  ! sd3
        p1=inputpar(7)          ! p1
        p2=inputpar(8)          ! p2
        rho=inputpar(9)         ! rho
        transition=0
        n=SIZE(states)
        DO i=1,n-1
            nodes(i)=(states(i+1)+states(i))/2.0_dp
        END DO
        DO i=1,n
            DO j=1,n
                IF (j .EQ. 1) THEN
                    transition(i,j)=alnorm((nodes(1)-rho*states(i)-m1)/s1,.false.)*p1 +      &
                                  & alnorm((nodes(1)-rho*states(i)-m2)/s2,.false.)*p2 +      &
                                  & alnorm((nodes(1)-rho*states(i)-m3)/s3,.false.)*(1.0_dp-p1-p2)
                    ELSE IF (j .EQ. n) THEN
                        transition(i,j)=alnorm((nodes(n-1)-rho*states(i)-m1)/s1,.true.)*p1 + &
                                      & alnorm((nodes(n-1)-rho*states(i)-m2)/s2,.true.)*p2 + &
                                      & alnorm((nodes(n-1)-rho*states(i)-m3)/s3,.true.)*(1.0_dp-p1-p2)    
                    ELSE
                        transition(i,j)=alnorm((nodes(j)   - rho*states(i)-m1)/s1,.false.)*p1 - &
                                      & alnorm((nodes(j-1) - rho*states(i)-m1)/s1,.false.)*p1 + &
                                      & alnorm((nodes(j)   - rho*states(i)-m2)/s2,.false.)*p2 - &
                                      & alnorm((nodes(j-1) - rho*states(i)-m2)/s2,.false.)*p2 + &
                                      & alnorm((nodes(j)   - rho*states(i)-m3)/s3,.false.)*(1.0_dp-p1-p2) - &
                                      & alnorm((nodes(j-1) - rho*states(i)-m3)/s3,.false.)*(1.0_dp-p1-p2)
                END IF                                              
            END DO
        END DO
!         print*,transition(1,:)           
!         print*,transition(2,:)           
!         print*,transition(3,:)
!         print*,states
!         print*,''      
    END SUBROUTINE TransitionOneStep
     
    FUNCTION RandomSimplex(d,MinP,MaxP)
        INTEGER, INTENT(IN)                 :: d        ! the number of dimensions
        REAL(DP), INTENT(IN)                :: MinP     ! the smallest realization of the process
        REAL(DP), INTENT(IN)                :: MaxP     ! the biggest realization of the process
        REAL(DP), DIMENSION(d+1,d)          :: RandomSimplex  ! the simplex
        ! This function generates a random grid of states
        ! The point of the grid are randomly drawn.
        ! A point can take any value between 1/2 of the minimum realization of the process
        ! up to 3/2 of the maximum realization of the process.
        INTEGER                             :: i,j        ! counter
        INTEGER                             :: info
        !
        DO i=1,d+1
            DO j=1,d
                RandomSimplex(i,j)=rand()*(MaxP-MinP)+MinP
            END DO
            ! sorting with lapack
            CALL dlasrt('I',d,RandomSimplex(i,:),info)
        END DO
    END FUNCTION RandomSimplex
      
    FUNCTION stationary(t)
        REAL(dp), DIMENSION(:,:), INTENT(IN)    ::  t
        REAL(dp), DIMENSION(size(t,1))          ::  stationary        
        !
        REAL(dp), DIMENSION(size(t,1))          ::  sOUT    
        REAL(dp), DIMENSION(size(t,1),size(t,1))::  temp,uOUT,vOUT
        INTEGER                                 ::  i,n,info
        REAL(dp), DIMENSION(5*size(t,1))        ::  WORK
        !
        n=size(t,1)
        temp=TRANSPOSE(t)
        DO i=1,n
            temp(i,i)=temp(i,i)-1.0_dp
        END DO
        CALL dgesvd ('N','A',n,n,temp,n,sOUT,uOUT,n,vOUT,n,WORK,5*n,info)
        stationary=vOUT(n,:)/sum(vOUT(n,:))
    END FUNCTION stationary
    
    FUNCTION TransitionStates2Targets(t,s)
        REAL(dp), DIMENSION(:,:), INTENT(IN)    ::  t
        REAL(dp), DIMENSION(:), INTENT(IN)      ::  s
        REAL(dp), DIMENSION(9)                  ::  TransitionStates2Targets        
        ! This function returns some characteristics of a markov
        ! process described by a transition matrix t and a vector of
        ! states s
        ! The characteristics that are computed are in order:
        ! The mean of the process
        ! The autocorrelation of the process
        ! The variance of the differences
        ! The skewness of the differences
        ! The kurtosis of the differences
        !
        REAL(dp), DIMENSION(size(s))          ::  stationary,sOUT    
        REAL(dp), DIMENSION(size(s),size(s))  ::  temp,uOUT,vOUT
        INTEGER                               ::  i,j,n,info
        REAL(dp), DIMENSION(size(s),size(s))  ::  pMatrix
        REAL(dp), DIMENSION(size(s),size(s))  ::  dMatrix
        REAL(dp), DIMENSION(size(s),size(s))  ::  mMatrix
        REAL(dp), DIMENSION(9)                ::  m
        REAL(dp), DIMENSION(5*size(s))        ::  WORK
        !
        n=size(s)
        ! The first thing we do here is computing the stationary distribution
        ! To do that we compute the singular value decomposition using lapack
        ! The singular value decomposition decomposes a matrix A in
        ! A=U*S*transpose(V)
        ! S is a diagonal matrix whose elements are positive
        ! if A is singular then at least one element of the diagonal of S
        ! is going to be zero (in numerical applications close to zero)
        ! also the row in V corresponding to a zero diagonal element of S
        ! is a base for the null space of A
        ! now, if A is the transpose of transition matrix minus the identity
        ! the null space is exactly the stationary distribution
        ! in fact p*T=p implies p(T-I)=0 implies transpose(T-I)*p'=0
        ! since the lapack algorithm returns the singular values
        ! in decreasing order, also the associated vectors are in descr.
        ! order. once we have a basis, normalizing it to one we get the 
        ! stationary distribution
        temp=TRANSPOSE(t)
        DO i=1,n
            temp(i,i)=temp(i,i)-1.0_dp
        END DO
        CALL dgesvd ('N','A',n,n,temp,n,sOUT,uOUT,n,vOUT,n,WORK,5*n,info)
        stationary=vOUT(n,:)/sum(vOUT(n,:))

        m(1)=sum(stationary*s)
        m(2)=sum((s-m(1))**2.0_dp*stationary)
        IF (m(2) .EQ. 0.0_dp) m(2)=EPSILON(1.0_dp)
        m(3)=sum((s-m(1))**3.0_dp*stationary)/m(2)**(3.0_dp/2.0_dp)
        m(4)=sum((s-m(1))**4.0_dp*stationary)/m(2)**2.0_dp
        DO i=1,n
            DO j=1,n
                pMatrix(i,j)=t(i,j)*stationary(i)
            END DO
        END DO
        DO i=1,n
            DO j=1,n
                mMatrix(i,j)=(s(j)-m(1))*(s(i)-m(1))/m(2)
            END DO
        END DO
        m(9)=sum(mMatrix*pMatrix)
        DO i=1,n
            DO j=1,n
                dMatrix(i,j)=s(j)-s(i)*m(9)
            END DO
        END DO
        m(5)=sum(dMatrix*pMatrix)
        m(6)=sum((dMatrix-m(5))**2.0*pMatrix)
        m(7)=sum((dMatrix-m(5))**3.0*pMatrix)/m(6)**(3.0_dp/2.0_dp)
        m(8)=sum((dMatrix-m(5))**4.0*pMatrix)/m(6)**2.0_dp
        TransitionStates2Targets=m
    END FUNCTION TransitionStates2Targets
    
    !*****************************************************************************
    !*****************************************************************************
    ! The following subroutines are taken from Numerical Recipes in Fortran 77: 
    ! The Art of Scientific Computing (1997) by William H. Press et al., 2nd Edition
    ! A useful and detailed exposition of these subroutines and how to 
    ! implement them in practice can be found in the book.
    !*****************************************************************************
    !*****************************************************************************
    
    
    SUBROUTINE amoeba(p,y,ftol,iter,inputpar,moments)
        ! multidimensional minimization by the downhill simplex method of Nelder and Mead.
    IMPLICIT NONE
    INTEGER, INTENT(OUT)                    :: iter
    REAL(dp), INTENT(IN)                    :: ftol
    REAL(dp), DIMENSION(:), INTENT(INOUT)   :: y ! "func" evaluated at the n vertices provided in "p"
    REAL(dp), DIMENSION(:,:), INTENT(INOUT) :: p ! vertices. If we have n vertices, then we must be
                                                 ! in n-1 dimensional space (we need one extra vertex
                                                 ! than dimensions. For each row, the n-1 vector
                                                 ! specifies the vertex
    REAL(dp), INTENT(IN)                    :: inputpar(9)
    REAL(dp), DIMENSION(9)                  :: moments
    INTEGER, PARAMETER :: ITMAX=1000
    REAL(dp), PARAMETER :: TINY=1.0e-10
    INTEGER :: ihi,ndim
    REAL(dp), DIMENSION(size(p,2)) :: psum

    call amoeba_private

    CONTAINS

    SUBROUTINE amoeba_private
        ! auxiliary function to amoeba.   
    IMPLICIT NONE
    INTEGER :: i,ilo,inhi
    REAL(dp) :: rtol,ysave,ytry,ytmp
    ndim=size(p,2)
    iter=0
    psum(:)=sum(p(:,:),dim=1)
    do
        ilo=iminloc(y(:))
        ihi=imaxloc(y(:))
        ytmp=y(ihi)
        y(ihi)=y(ilo)
        inhi=imaxloc(y(:))
        y(ihi)=ytmp
        rtol=2.0_dp*abs(y(ihi)-y(ilo))/(abs(y(ihi))+abs(y(ilo))+TINY)
        if (rtol < ftol) then
            call swap(y(1),y(ilo))
            call swap_v(p(1,:),p(ilo,:))
            RETURN
        end if
        if (iter >= ITMAX) then
!             print*,'too many iterations, the program will continue with values found in the last iteration'
            call swap(y(1),y(ilo))
            call swap_v(p(1,:),p(ilo,:))
            return
        end if
        ytry=amotry(-1.0_dp)
        iter=iter+1
        if (ytry <= y(ilo)) then
            ytry=amotry(2.0_dp)
            iter=iter+1
        else if (ytry >= y(inhi)) then
            ysave=y(ihi)
            ytry=amotry(0.5_dp)
            iter=iter+1
            if (ytry >= ysave) then
                p(:,:)=0.5_dp*(p(:,:)+spread(p(ilo,:),1,size(p,1)))
                do i=1,ndim+1
                    if (i /= ilo) y(i)=ErrorFunction(inputpar,moments,p(i,:))
                end do
                iter=iter+ndim
                psum(:)=sum(p(:,:),dim=1)
            end if
        end if
    end do
    END SUBROUTINE amoeba_private

    FUNCTION amotry(fac)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: fac
    REAL(dp) :: amotry
    REAL(dp) :: fac1,fac2,ytry
    REAL(dp), DIMENSION(size(p,2)) :: ptry
    fac1=(1.0_dp-fac)/ndim
    fac2=fac1-fac
    ptry(:)=psum(:)*fac1-p(ihi,:)*fac2
    ytry=ErrorFunction(inputpar,moments,ptry)
    if (ytry < y(ihi)) then
        y(ihi)=ytry
        psum(:)=psum(:)-p(ihi,:)+ptry(:)
        p(ihi,:)=ptry(:)
    end if
    amotry=ytry
    END FUNCTION amotry
    END SUBROUTINE amoeba
    
    FUNCTION iminloc(arr)
        !location of array minimum as an integer
        REAL(DP), DIMENSION(:), INTENT(IN)  :: arr
        INTEGER, DIMENSION(1)               :: imin
        INTEGER                             :: iminloc
        imin=MINLOC(arr(:))
        iminloc=imin(1)
    END FUNCTION iminloc
        
    FUNCTION imaxloc(arr)
        !location of array maximum as an integer
        REAL(DP), DIMENSION(:), INTENT(IN)  :: arr
        INTEGER                             :: imaxloc
        INTEGER, DIMENSION(1)               :: imax
        imax=MAXLOC(arr(:))
        imaxloc=imax(1)
    END FUNCTION imaxloc
        
    SUBROUTINE swap(a,b)
        REAL(dp), INTENT(INOUT) :: a,b
        REAL(dp) :: dum
        dum=a
        a=b
        b=dum
    END SUBROUTINE swap
        
    SUBROUTINE swap_v(a,b)
        REAL(dp), DIMENSION(:), INTENT(INOUT) :: a,b
        REAL(dp), DIMENSION(SIZE(a)) :: dum
        dum=a
        a=b
        b=dum
    END SUBROUTINE swap_v
    
    !*****************************************************************************
    !*****************************************************************************
    
    FUNCTION ProcessToMoments(process)
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(IN) ::  process
    REAL(dp), DIMENSION(9)             ::  ProcessToMoments
    REAL(dp), DIMENSION(size(process)) ::  e
    ! Returns the first 4 centered moments of the process
    REAL(dp)                           ::  t
    !
    t=REAL(SIZE(process))
    ProcessToMoments(9)=autocorr1(process)
    ProcessToMoments(1)=SUM(process)/t
    ProcessToMoments(2)=SUM((process-ProcessToMoments(1))**2.0_dp)/t
    ProcessToMoments(3)=SUM((process-ProcessToMoments(1))**3.0_dp)/t/ProcessToMoments(2)**(3.0_dp/2.0_dp)
    ProcessToMoments(4)=SUM((process-ProcessToMoments(1))**4.0_dp)/t/ProcessToMoments(2)**(2.0_dp)
    e=process*ProcessToMoments(9)
    e(2:INT(t))=e(1:INT(t)-1)
    e=process-e
    ProcessToMoments(5)=SUM(e)/t
    ProcessToMoments(6)=SUM((e-ProcessToMoments(5))**2.0_dp)/t
    ProcessToMoments(7)=SUM((e-ProcessToMoments(5))**3.0_dp)/t/ProcessToMoments(6)**(3.0_dp/2.0_dp)
    ProcessToMoments(8)=SUM((e-ProcessToMoments(5))**4.0_dp)/t/ProcessToMoments(6)**(2.0_dp)
    END FUNCTION ProcessToMoments
    
    FUNCTION mom1(process)
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(IN)  ::  process
    REAL(dp)                            ::  mom1
    !
    mom1=SUM(process)/REAL(SIZE(process))
    END FUNCTION mom1
    
    FUNCTION mom2(process)
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(IN) ::  process
    REAL(dp)                           ::  mom2
    !
    mom2=SUM((process-mom1(process))**2.0_dp)/REAL(size(process))
    END FUNCTION mom2
    
    FUNCTION mom3(process)
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(IN) ::  process
    REAL(dp)                           ::  mom3
    !
    mom3=SUM((process-mom1(process))**3.0_dp)/REAL(SIZE(process))/mom2(process)**(3.0_dp/2.0_dp)
    END FUNCTION mom3
    
    FUNCTION mom4(process)
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(IN) ::  process
    REAL(dp)                           ::  mom4
    !
    mom4=SUM((process-mom1(process))**4.0_dp)/REAL(SIZE(process))/mom2(process)**2.0_dp
    END FUNCTION mom4
    
    FUNCTION autocorr1(process)
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(IN) ::  process
    REAL(dp)                           ::  autocorr1
    REAL(dp)                           ::  temp
    INTEGER                            ::  i,t
    !
    t=size(process)
    temp=0.0_dp
    DO i=1,t-1
        temp=process(i)*process(i+1)+temp
    END DO
    temp=temp/real(t-1)
    autocorr1=(temp-mom1(process)**2.0_dp)/mom2(process)
    END FUNCTION autocorr1
    
    FUNCTION Diff(process)
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(IN)      ::  process
    REAL(dp), DIMENSION(size(process)-1)    ::  Diff
    ! Returns the first differences
    INTEGER                             ::  i
    DO i=1,size(process)-1
        Diff(i)=process(i+1)-process(i)
    END DO
    END FUNCTION Diff
    
    FUNCTION linspace(a,b,num)
        IMPLICIT NONE
        REAL(DP), INTENT(IN)                :: a,b
        INTEGER, INTENT(IN)                 :: num
        REAL(dp), DIMENSION(num)             :: linspace
        ! linspace generates a linearly spaced grid in the interval [a,b]
        ! with the number of points being determined by num. Clearly a < b is required and
        ! num > 0. The output is x, the linearly spaced grid
        INTEGER  :: i
        linspace = (/ (a+(real(i,dp)-1.0_dp)*b/(real(num,dp)-1.0), i = 1, num) /)
    END FUNCTION linspace
    
    FUNCTION randnorm()
        IMPLICIT NONE
        REAL(dp)    :: randnorm
        ! returns a normally distributed number with zero mean and unit variance
        REAL(dp)    :: rsq,v1,v2
        !
        DO
            v1=rand()
            v2=rand()
            v1=2.0_dp*v1-1.0_dp
            v2=2.0_dp*v2-1.0_dp
            rsq=v1**2+v2**2
            IF (rsq > 0.0 .and. rsq < 1.0) EXIT
        END DO
        rsq=sqrt(-2.0_dp*log(rsq)/rsq)
        randnorm=v1*rsq
    END FUNCTION randnorm
    
    function alnorm ( x, upper )
    !*****************************************************************************80
    !
    !! ALNORM computes the cumulative density of the standard normal distribution.
    !
    !  Licensing:
    !
    !    This code is distributed under the GNU LGPL license.
    !
    !  Modified:
    !
    !    13 January 2008
    !
    !  Author:
    !
    !    Original FORTRAN77 version by David Hill.
    !    FORTRAN90 version by John Burkardt.
    !
    !  Reference:
    !
    !    David Hill,
    !    Algorithm AS 66:
    !    The Normal Integral,
    !    Applied Statistics,
    !    Volume 22, Number 3, 1973, pages 424-427.
    !
    !  Parameters:
    !
    !    Input, REAL(dp) X, is one endpoint of the semi-infinite interval
    !    over which the integration takes place.
    !
    !    Input, logical UPPER, determines whether the upper or lower
    !    interval is to be integrated:
    !    .TRUE.  => integrate from X to + Infinity;
    !    .FALSE. => integrate from - Infinity to X.
    !
    !    Output, REAL(dp) ALNORM, the integral of the standard normal
    !    distribution over the desired interval.
    !
      implicit none

      REAL(dp), parameter :: a1 = 5.75885480458D+00
      REAL(dp), parameter :: a2 = 2.62433121679D+00
      REAL(dp), parameter :: a3 = 5.92885724438D+00
      REAL(dp) alnorm
      REAL(dp), parameter :: b1 = -29.8213557807D+00
      REAL(dp), parameter :: b2 = 48.6959930692D+00
      REAL(dp), parameter :: c1 = -0.000000038052D+00
      REAL(dp), parameter :: c2 = 0.000398064794D+00
      REAL(dp), parameter :: c3 = -0.151679116635D+00
      REAL(dp), parameter :: c4 = 4.8385912808D+00
      REAL(dp), parameter :: c5 = 0.742380924027D+00
      REAL(dp), parameter :: c6 = 3.99019417011D+00
      REAL(dp), parameter :: con = 1.28D+00
      REAL(dp), parameter :: d1 = 1.00000615302D+00
      REAL(dp), parameter :: d2 = 1.98615381364D+00
      REAL(dp), parameter :: d3 = 5.29330324926D+00
      REAL(dp), parameter :: d4 = -15.1508972451D+00
      REAL(dp), parameter :: d5 = 30.789933034D+00
      REAL(dp), parameter :: ltone = 7.0D+00
      REAL(dp), parameter :: p = 0.398942280444D+00
      REAL(dp), parameter :: q = 0.39990348504D+00
      REAL(dp), parameter :: r = 0.398942280385D+00
      logical up
      logical upper
      REAL(dp), parameter :: utzero = 18.66D+00
      REAL(dp) x
      REAL(dp) y
      REAL(dp) z

      up = upper
      z = x

      if ( z < 0.0D+00 ) then
        up = .not. up
        z = - z
      end if

      if ( ltone < z .and. ( ( .not. up ) .or. utzero < z ) ) then

        if ( up ) then
          alnorm = 0.0D+00
        else
          alnorm = 1.0D+00
        end if

        return

      end if

      y = 0.5D+00 * z * z

      if ( z <= con ) then

        alnorm = 0.5D+00 - z * ( p - q * y &
          / ( y + a1 + b1 &
          / ( y + a2 + b2 &
          / ( y + a3 ))))

      else

        alnorm = r * exp ( - y ) &
          / ( z + c1 + d1 &
          / ( z + c2 + d2 &
          / ( z + c3 + d3 &
          / ( z + c4 + d4 &
          / ( z + c5 + d5 &
          / ( z + c6 ))))))

      end if

      if ( .not. up ) then
        alnorm = 1.0D+00 - alnorm
      end if

      return
    end
    
end module tools_module
